
let _dataVisualizationConfigRelativeFilePath = 'desite.modules/dataVisualization.module/dataVisualizationConfig.json';
let _dataVisualizationConfigFilePath = '';
let _dataVisualizationConfigFileFound = false;

let _dataVisualizationConfigTemplate = {
    "Bookmark": "",
    "ChartType": "",
    "Properties": {
        "Axis": {
            "Name": "",
            "Type": ""
        },
        "Value": {
            "Name": "",
            "Type": ""
        },
        "Legend": {
            "Name": "",
            "Type": ""
        }
    },
    "PropertyValuesPresentedAs": "",
    "ChartTitle": "Chart Title"
}

let _dataVisualizationConfig = {};
let _configCurrent = {};
let _chartSettingsComplete = false;

let _visibilityData = {};

let _categoryAxisObject = {
    type: 'category',
    name: '',
    nameLocation: '', // for X-Axis: middle, for Y-Axis: end
    nameGap: 0,
    data: []
}

let _valueAxisObject = {
    type: 'value',
    name: '',
    nameLocation: '', // for X-Axis: middle, for Y-Axis: end
    nameGap: 20       // for X-Axis: 20, for Y-Axis: 20
}

let _dataZoomBarChart =
[
    {
        show: true,
        xAxisIndex: 0,
        filterMode: 'none',
        bottom: '8%'
    },
    {
        show: true,
        yAxisIndex: 0,
        filterMode: 'filter',
        right: "0%"
    }    
]

let _dataZoomColumnChart =
[
    {
        show: true,
        xAxisIndex: 0,
        filterMode: 'filter',
        bottom: '8%'
    },
    {
        show: true,
        yAxisIndex: 0,
        filterMode: 'none',
        right: "0%"
    }
]

let _echart;
// Specify the configuration items and data for the chart
let _echartOption = {
    tooltip: {
        trigger: 'item',
        axisPointer: {
            // When axis is used to trigger tooltip:
            type: 'shadow' // 'shadow' as default; can also be 'line' or 'shadow'
        }        
    },
    toolbox: {
        show: true,
        feature: {
            magicType: { type: ['stack'] }                            
        }
    },
    label: {
        show: true,
        formatter: function (param) {
            return param.data == 0 ? '' : param.data;
        }
    },
    legend: {
        type: 'scroll',
        orient: 'horizontal',
        top: 40
    },
    grid: [
        {top: '20%', left: '8%', right: '4%', bottom: '15%',  containLabel: true},
        {top: '20%', left: '5%', bottom: '15%',  containLabel: true},
        {top: '20%', left: '8%', right: '4%', bottom: '5%', containLabel: false}
    ],
    xAxis: [
        {
            type: 'value',
            name: 'X-Axis Name',
            nameLocation: 'middle',
            nameGap: 20
        },
        {
            gridIndex: 1,
            type: 'value',
            axisLine: {show: false}
        },
        {
            gridIndex: 2,
            type: 'category',
            name: 'Name2',
            nameLocation: 'middle',
            axisLine: {show: false}
        }
    ],
    yAxis: [
        {
            type: 'category',
            name: 'Y-Axis Name',
            nameLocation: 'end',
            nameGap: 20,
            data: []
        },
        {
            gridIndex: 1,
            name: 'Name1',
            nameLocation: 'middle', nameGap: 0, nameRotate: 90,
            type: 'category',
            data: [],
            axisLine: {show: false}
        },
        {
            gridIndex: 2,
            type: 'category',
            data: [],
            axisLine: {show: false}
        }
    ],    
    dataZoom: [],
    series: [
        {
            name: '',
            type: 'bar',
            stack: 'total',
            label: {
                show: true,
                position: 'right'
            },
            emphasis: {
                focus: 'series'
            },
            data: []
        }        
    ]
};

let _visibleGeometricalObjects = [];
let _echartClicked = false;
let _showEvaluatedObjectsOnlyClicked = false;

let _progressObject = {
    text: false,
    bar: false,
    label: true,
    closeButton: true,
    title: "Progress"
}

addEventListener('desiteload', async function(event)
{
    _dataVisualizationConfig = await readOrCreateDataVisualizationConfigFile();
    console.log(JSON.stringify(_dataVisualizationConfig, null, 2))
    _configCurrent = JSON.parse(JSON.stringify(_dataVisualizationConfigTemplate));
    await createBookmarkDropdownMenu();
    await setDropdownMenuOptionFromConfig('dropdownPropertyForAxis');
    await setDropdownMenuOptionFromConfig('dropdownPropertyForValue');
    await setDropdownMenuOptionFromConfig('dropdownPropertyForLegend');
    await setDropdownChartTypeFromConfig();
    await createPropertyValuesPresentedAsDropdownMenu([]);
    await setDropdownPropertyValuesPresentedAsFromConfig();
    await setChartTitleFromConfig();
    $('#showEvaluatedObjectsOnlyButton').prop('disabled', true);
   
    desiteAPI.visibilityChanged.connect(onVisibilityChanged);
    // Initialize the echarts instance based on the prepared dom
    _echart = echarts.init(document.getElementById('chartDiv'));

    _echart.on('click', async function (params) {
        _echartClicked = true;
        //console.log(Object.keys(params)); 
        //console.log(params.dataIndex); 
        //console.log(params.name); // data name
        //console.log(params.seriesIndex);  
        //console.log(params.seriesName);  
        const axisPropertyValue = params.name;
        const legendPropertyValue = params.seriesName;
        if (Object.keys(_visibilityData).indexOf('Legend') == -1)
        {
            await isolateElements(_visibilityData.objectsByPropertyValues[axisPropertyValue]);
            $("#visibleObjectsInfo").html(_visibilityData.objectsByPropertyValues[axisPropertyValue].length  + ' of '  +  _visibleGeometricalObjects.length + ' objects visible');
        } 
        else
        {
            await isolateElements(_visibilityData.objectsByPropertyValues[axisPropertyValue][legendPropertyValue]);
            $("#visibleObjectsInfo").html(_visibilityData.objectsByPropertyValues[axisPropertyValue][legendPropertyValue].length  + ' of '  +  _visibleGeometricalObjects.length + ' objects visible');
        }  
        setTimeout(() => {
            _echartClicked = false;
        }, 100);            
    }); 
});            

async function readOrCreateDataVisualizationConfigFile()
{
    let config = {};
    const projectPath = await desiteAPI.getProjectDirectory();
    _dataVisualizationConfigFilePath = projectPath + '/' + _dataVisualizationConfigRelativeFilePath;
    const fileExists = await desiteAPI.checkIfFileExists(_dataVisualizationConfigFilePath);
    if (fileExists) 
    {       
        _dataVisualizationConfigFileFound = true;
        config = JSON.parse(await desiteAPI.readTextFileAsString(_dataVisualizationConfigFilePath, "UTF-8"));
    }
    else
    {
        _dataVisualizationConfigFileFound = false;
        alert(`Couldn't find config file '${_dataVisualizationConfigFilePath}'.\nA new file will be created.`);
        const returns = await desiteAPI.writeFile(_dataVisualizationConfigFilePath,  JSON.stringify(_dataVisualizationConfigTemplate, undefined, 3), "UTF-8");
        console.log(returns);
        config = JSON.parse(JSON.stringify(_dataVisualizationConfigTemplate));
    }        
    return config;
}

async function isolateElements(elementIDs)
{
    await desiteAPI.setAllElementsVisible(false, 'geometry');
    await desiteAPI.setElementsVisible(true, elementIDs);
}

async function onVisibilityChanged()
{
    if (!_echartClicked  && !_showEvaluatedObjectsOnlyClicked)
    {
        $("#visibleObjectsInfo").html('');
    }
     
}

window.addEventListener('resize', async function () {
    _echart.resize();
});


async function sumValues(objects, property)
{

    let sum = 0;
    for (let objId of objects)
    {
        sum += await desiteAPI.getPropertyValue(objId, property.Name, property.Type);
    }
    return sum;
}

async function displayChart()
{
    const visibleObjects = await desiteAPI.getVisibleElements("geometry");
    _visibleGeometricalObjects = await desiteAPI.filterByProperty(visibleObjects, 'cpCountTriangles', 'xs:long', '>0');

    if (_visibleGeometricalObjects.length == 0)
    {
        $("#chartDiv").hide();
        $("#evaluatedObjectsInfo").html('');
        $("#visibleObjectsInfo").html('');
        $("#ObjectsInfoDiv").hide();
        
        setTimeout(() => {
            alert('No geometrical objects are visible!');           
        }, 100);
        return;
    }

    await checkChartSettings();
       
    if (!_chartSettingsComplete)
    {
        //$("#chartDiv").hide();
        return;
    }
    else if (_configCurrent.Properties.Legend.Name == '')
    {
        desiteMD.showProgress(_progressObject);
        desiteMD.showProgressMessageLabel('Getting data for chart...');   

        _visibilityData = await createDataForChartWithoutLegend(_visibleGeometricalObjects);

        if (_visibilityData.objects.length == 0)
        {
            $("#chartDiv").hide();
            $("#evaluatedObjectsInfo").html('');
            $("#visibleObjectsInfo").html('');
            $("#ObjectsInfoDiv").hide();
           
            setTimeout(function(){
                alert("No visible geometrical objects have a value in the property '" + _configCurrent.Properties.Value.Name + "'!");           
            },100);
            return;
        }

        const property = _configCurrent.Properties.Axis;
        console.log(JSON.stringify(property, null, 2))        
       
        const propertyValues = Object.keys(_visibilityData.objectsByPropertyValues);
        
        let seriesData = [];        
        
        if (_configCurrent.PropertyValuesPresentedAs == 'Count')
        {
            if (_configCurrent.ChartType == 'BarChart' ||  _configCurrent.ChartType == 'ColumnChart')
            {
                for (let value of propertyValues)
                {
                    seriesData.push(_visibilityData.objectsByPropertyValues[value].length);
                } 
            }
            else /* _configCurrent.ChartType == 'BarChartPercentage' ||  _configCurrent.ChartType == 'ColumnChartPercentage' */
            {
                console.log('_visibilityData.numberOfObjects: ' + _visibilityData.objects.length)
                for (let value of propertyValues)
                {
                    let valueInPercent = _visibilityData.objectsByPropertyValues[value].length / _visibilityData.objects.length * 100;
                    valueInPercent = +valueInPercent.toFixed(4);
                    seriesData.push(valueInPercent);
                } 
            }
        }
        else /* Sum */
        {
            if (_configCurrent.ChartType == 'BarChart' ||  _configCurrent.ChartType == 'ColumnChart')
            {
                for (let value of propertyValues)
                {
                    let sum = await sumValues(_visibilityData.objectsByPropertyValues[value], _configCurrent.Properties.Value);
                    sum = +sum.toFixed(4);
                    seriesData.push(sum);
                } 
            }
            else /* _configCurrent.ChartType == 'BarChartPercentage' ||  _configCurrent.ChartType == 'ColumnChartPercentage' */
            {
                const sumValuesAllObjects = await sumValues(_visibilityData.objects, _configCurrent.Properties.Value);
                if (sumValuesAllObjects != 0)
                {
                    for (let value of propertyValues)
                    {
                        let sumValueInPercent = await sumValues(_visibilityData.objectsByPropertyValues[value], _configCurrent.Properties.Value) / sumValuesAllObjects * 100;;
                        sumValueInPercent = +sumValueInPercent.toFixed(4);
                        seriesData.push(sumValueInPercent);
                    } 
                }
                else
                {
                    for (let value of propertyValues)
                    {
                        seriesData.push(0);
                    } 
                }                
            }
        }
        // Set echart option
        //_echartOption.legend = {};
        if(_configCurrent.ChartType == 'BarChart' ||  _configCurrent.ChartType == 'BarChartPercentage')
        {
            _echartOption.yAxis[0] = JSON.parse(JSON.stringify(_categoryAxisObject));
            _echartOption.yAxis[1].name = _configCurrent.Properties.Axis.Name;
            _echartOption.yAxis[0].data = propertyValues;
            _echartOption.yAxis[0].data = propertyValues;
            _echartOption.xAxis[0] = JSON.parse(JSON.stringify(_valueAxisObject));
            _echartOption.dataZoom = _dataZoomBarChart;

            if (_configCurrent.ChartType == 'BarChart')
            {
                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.xAxis[2].name = 'Count objects'; // ?
                }
                else /* Sum */
                {
                    _echartOption.xAxis[2].name = 'Sum of ' + _configCurrent.Properties.Value.Name;
                }
            }
            else /* BarChartPercentage */
            {
                _echartOption.xAxis[0].max = 100; //?
                _echartOption.xAxis[0].axisLabel = { formatter: '{value}%' }

                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.xAxis[2].name = 'Count objects (%) ';                   
                }
                else /* Sum */
                {
                    _echartOption.xAxis[2].name =  'Sum of ' + _configCurrent.Properties.Value.Name + ' (%) ';
                }                
            }           
        }
        else /* _configCurrent.ChartType == 'ColumnChart' ||  _configCurrent.ChartType == 'ColumnChartPercentage' */
        {
            _echartOption.yAxis[0] = JSON.parse(JSON.stringify(_valueAxisObject));
            _echartOption.xAxis[0] = JSON.parse(JSON.stringify(_categoryAxisObject));
            _echartOption.xAxis[2].name = _configCurrent.Properties.Axis.Name;
            _echartOption.xAxis[0].data = propertyValues;
            _echartOption.xAxis[0].axisLabel = {
                rotate: 20
            }            
            _echartOption.dataZoom = _dataZoomColumnChart;
            if (_configCurrent.ChartType == 'ColumnChart')
            {
                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.yAxis[1].name = 'Count objects';
                }
                else /* Sum */
                {
                    _echartOption.yAxis[1].name = 'Sum of ' + _configCurrent.Properties.Value.Name;
                }
            }
            else /* ColumnChartPercentage */
            {
                _echartOption.yAxis[0].max = 100; //?
                _echartOption.yAxis[0].axisLabel = { formatter: '{value}%' }

                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.yAxis[1].name = 'Count objects (%) ';                   
                }
                else /* Sum */
                {
                    _echartOption.yAxis[1].name =  'Sum of ' + _configCurrent.Properties.Value.Name + ' (%) ';
                }                
            }
        }
        _echartOption.series = [
            {
                name: property.Name,
                type: 'bar',
                stack: 'total',
                label: {
                    show: true
                },
                emphasis: {
                    focus: 'series'
                },
                data: seriesData
            }
        ];
        _echartOption.toolbox.feature.magicType = {};
        console.log(_echartOption.series.length);
        console.log(JSON.stringify(_echartOption, null, 2));
        desiteMD.hideProgress();        
    }
    else /* Property for Legend is not empty */
    {   
        desiteMD.showProgress(_progressObject);
        desiteMD.showProgressMessageLabel('Getting data for chart...');

        _visibilityData = await createDataForChartWithLegend(_visibleGeometricalObjects);
        //console.log('_visibilityData:')
        //console.log(JSON.stringify(_visibilityData, null, 2))

        if (_visibilityData.objects.length == 0)
        {
            $("#chartDiv").hide();
            $("#evaluatedObjectsInfo").html('');
            $("#visibleObjectsInfo").html('');
            $("#ObjectsInfoDiv").hide();

            setTimeout(function(){
                alert("No visible geometrical objects have a value in the property '" + _configCurrent.Properties.Value.Name + "'!");           
            },100);
            return;
        }

        const property = _configCurrent.Properties.Axis;
        console.log(JSON.stringify(property, null, 2))

        if ((_configCurrent.ChartType == 'BarChartPercentage' ||  _configCurrent.ChartType == 'ColumnChartPercentage') && _configCurrent.PropertyValuesPresentedAs == 'Sum')
        {
            // calculate and add sum values to _visibilityData.Axis.sumsByPropertyValues
            for (let axisPropValue of _visibilityData.Axis.propertyValues)
            {
                _visibilityData.Axis.sumsByPropertyValues[axisPropValue] = await sumValues(_visibilityData.Axis.objectsByPropertyValues[axisPropValue], _configCurrent.Properties.Value);
            }
        }

        let seriesList = [];

        for (let legendPropValue of _visibilityData.Legend.propertyValues)
        {
            let series = {
                name: legendPropValue,
                type: 'bar',
                stack: 'total',
                label: {
                    show: true
                },
                emphasis: {
                    focus: 'series'
                },
                data: []
            }
            
            for (let axisPropValue of _visibilityData.Axis.propertyValues)
            {
                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    if (_configCurrent.ChartType == 'BarChart' ||  _configCurrent.ChartType == 'ColumnChart')
                    {
                        series.data.push(_visibilityData.objectsByPropertyValues[axisPropValue][legendPropValue].length)
                    }
                    else /* _configCurrent.ChartType == 'BarChartPercentage' ||  _configCurrent.ChartType == 'ColumnChartPercentage' */
                    {
                        let valueInPercent = _visibilityData.objectsByPropertyValues[axisPropValue][legendPropValue].length / _visibilityData.Axis.objectsByPropertyValues[axisPropValue].length * 100;
                        valueInPercent = +valueInPercent.toFixed(4);
                        series.data.push(valueInPercent);
                    }
                }
                else /* Sum */
                {
                    let sum = await sumValues(_visibilityData.objectsByPropertyValues[axisPropValue][legendPropValue], _configCurrent.Properties.Value);
                    if (_configCurrent.ChartType == 'BarChart' ||  _configCurrent.ChartType == 'ColumnChart')
                    {                        
                        sum = +sum.toFixed(4);
                        series.data.push(sum);
                    }
                    else /* _configCurrent.ChartType == 'BarChartPercentage' ||  _configCurrent.ChartType == 'ColumnChartPercentage' */
                    {
                        if (_visibilityData.Axis.sumsByPropertyValues[axisPropValue] != 0)
                        {
                            let sumValueInPercent = sum / _visibilityData.Axis.sumsByPropertyValues[axisPropValue] * 100;
                            sumValueInPercent = +sumValueInPercent.toFixed(4);
                            series.data.push(sumValueInPercent);
                        }
                        else
                        {
                            series.data.push(0);
                        }
                    }
                }
            }
            seriesList.push(series);    
        }
        console.log('seriesList:')
        console.log(JSON.stringify(seriesList, null, 2)) 
        
        if(_configCurrent.ChartType == 'BarChart' ||  _configCurrent.ChartType == 'BarChartPercentage')
        {
            _echartOption.yAxis[0] = JSON.parse(JSON.stringify(_categoryAxisObject));
            _echartOption.yAxis[1].name = _configCurrent.Properties.Axis.Name;
            _echartOption.yAxis[0].data = _visibilityData.Axis.propertyValues;

            _echartOption.xAxis[0] = JSON.parse(JSON.stringify(_valueAxisObject));
            _echartOption.dataZoom = _dataZoomBarChart;   

            if (_configCurrent.ChartType == 'BarChart')
            {
                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.xAxis[2].name = 'Count objects'; // ?
                }
                else /* Sum */
                {
                    _echartOption.xAxis[2].name = 'Sum of ' + _configCurrent.Properties.Value.Name;
                }
            }
            else /* BarChartPercentage */
            {
                _echartOption.xAxis[0].max = 100.1; //?
                _echartOption.xAxis[0].axisLabel = { formatter: '{value}%' }

                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.xAxis[2].name = 'Count objects (%) ';                     
                }
                else /* Sum */
                {
                    _echartOption.xAxis[2].name =  'Sum of ' + _configCurrent.Properties.Value.Name + ' (%) ';
                }                
            }
        }
        else /* _configCurrent.ChartType == 'ColumnChart' ||  _configCurrent.ChartType == 'ColumnChartPercentage' */
        {
            _echartOption.yAxis[0] = JSON.parse(JSON.stringify(_valueAxisObject));
            _echartOption.xAxis[0] = JSON.parse(JSON.stringify(_categoryAxisObject));
            _echartOption.xAxis[2].name = _configCurrent.Properties.Axis.Name;
            _echartOption.xAxis[0].data = _visibilityData.Axis.propertyValues;
            _echartOption.xAxis[0].axisLabel = {
                rotate: 20
            }            
            _echartOption.dataZoom = _dataZoomColumnChart;

            if (_configCurrent.ChartType == 'ColumnChart')
            {
                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.yAxis[1].name = 'Count objects';
                }
                else /* Sum */
                {
                    _echartOption.yAxis[1].name = 'Sum of ' + _configCurrent.Properties.Value.Name;
                }
            }
            else /* ColumnChartPercentage */
            {
                _echartOption.yAxis[0].max = 100.1; //?
                _echartOption.yAxis[0].axisLabel = { formatter: '{value}%' }

                if (_configCurrent.PropertyValuesPresentedAs == 'Count')
                {
                    _echartOption.yAxis[1].name = 'Count objects (%) ';                   
                }
                else /* Sum */
                {
                    _echartOption.yAxis[1].name =  'Sum of ' + _configCurrent.Properties.Value.Name + ' (%) ';
                }                
            }
        }
        _echartOption.series = seriesList;
        _echartOption.toolbox.feature.magicType = { type: ['stack'] };

        desiteMD.hideProgress();
    }
    $("#chartDiv").show();
    _echart.resize({
        width: 'auto',
        height: 'auto'
    });
    $("#evaluatedObjectsInfo").html(_visibleGeometricalObjects.length + ' objects evaluated');
    $("#ObjectsInfoDiv").show();
    _echart.setOption(_echartOption, true);
    $('#showEvaluatedObjectsOnlyButton').prop('disabled', false); 
    
    if(_configCurrent.Properties.Legend.Name != '')
    {
        await createAndShowColorSchemeForLegendProperty()
    }
}

async function createAndShowColorSchemeForLegendProperty()
{
    let intervalSize = 1.0;
    await desiteAPI.colorCodeObjects(_visibleGeometricalObjects, _configCurrent.Properties.Legend.Name, _configCurrent.Properties.Legend.Type, intervalSize);
    let colorSchemaName = "Data Visualization: " + _configCurrent.Properties.Legend.Name;
    let colorSchemaOptions =
    {
        legend: true
    };        
    await desiteAPI.createColorSchemaFromCurrentMaterials(colorSchemaName, true, colorSchemaOptions);        
    await desiteAPI.activateColorSchemaByName(colorSchemaName);
}

async function onShowEvaluatedObjectsOnlyClicked()
{
    _showEvaluatedObjectsOnlyClicked = true;
    await isolateElements(_visibleGeometricalObjects);
    $("#visibleObjectsInfo").html(_visibleGeometricalObjects.length  + ' of '  + _visibleGeometricalObjects.length + ' objects visible');
    setTimeout(() => {
        _showEvaluatedObjectsOnlyClicked = false
    }, 100);
}

function applyObjectFilter(objects, filter)
{
    let result = objects;
    for (let i in filter)
    {
        const filterElement = filter[i];
        result = desiteAPI.filterByProperty(result, filterElement.propertyName, filterElement.propertyDataType, filterElement.filterPattern, filterElement.checkInherited, filterElement.domains);
    }
    return result;
}

async function filterObjectsByProperty(objects, property, value)
{
    let result;
    if (value == '(empty)')
    {
        result = await desiteAPI.filterByProperty(objects, property.propertyName, property.propertyDataType, '!*', false, 'geometry');
    }
    else
    {
        result = await desiteAPI.filterByProperty(objects, property.propertyName, property.propertyDataType, '"' + value + '"', false, 'geometry');
    }
    return result
}
   
async function createDataForChartWithoutLegend(objects)
{
    const data =
    {
        property: {}, 
        objects: [],        
        objectsByPropertyValues: {} // Objects grouped by the values of Property for Axis
    }

    data.property = JSON.parse(JSON.stringify(_configCurrent.Properties.Axis));

    if (_configCurrent.Properties.Axis.Name != _configCurrent.Properties.Value.Name)
    {
        data.objects = await desiteAPI.filterByProperty(objects, _configCurrent.Properties.Value.Name, _configCurrent.Properties.Value.Type, '*');
    }
    else
    {
        data.objects = objects;
    }
    
    const values = await desiteAPI.getPropertyValuesByObjectList(_configCurrent.Properties.Axis.Name, _configCurrent.Properties.Axis.Type, true, data.objects, 0, 'geometry', true);
    console.log(values);
    
    for (let i in values)
    {
        const value = values[i];
        if (value == null)
        {
            data.objectsByPropertyValues['(empty)'] = [];
        }
        else
        {
            data.objectsByPropertyValues[value] = [];
        }
    }
    
    for (let i in data.objects)
    {
        const objId = data.objects[i];
        const propValue = await desiteAPI.getPropertyValue(objId, _configCurrent.Properties.Axis.Name, _configCurrent.Properties.Axis.Type);
        if (propValue != null)
        {
            data.objectsByPropertyValues[propValue].push(objId);
        }
        else
        {
            data.objectsByPropertyValues['(empty)'].push(objId);
        }
    }    
    return data;
}

async function createDataForChartWithLegend(objects)
{    
    const data =
    {
        objects: [],
        objectsByPropertyValues: {},  // Objects grouped by the values of Property for Axis and Property for Legend
        Axis: { 
            propertyValues: [],
            objectsByPropertyValues: {},  
            sumsByPropertyValues: {}      
        },
        Legend: {
            propertyValues: []
        }
    }    

    if (_configCurrent.Properties.Axis.Name != _configCurrent.Properties.Value.Name)
    {
        data.objects = await desiteAPI.filterByProperty(objects, _configCurrent.Properties.Value.Name, _configCurrent.Properties.Value.Type, '*');
    }
    else
    {
        data.objects = objects;
    }

    const valuesOfAxisProp = await desiteAPI.getPropertyValuesByObjectList(_configCurrent.Properties.Axis.Name, _configCurrent.Properties.Axis.Type, true, data.objects, 0, 'geometry', true);
    const valuesOfLegendProp = await desiteAPI.getPropertyValuesByObjectList(_configCurrent.Properties.Legend.Name, _configCurrent.Properties.Legend.Type, true, data.objects, 0, 'geometry', true);
    console.log('valuesOfAxisProp:' + JSON.stringify(valuesOfAxisProp, null, 2));
    console.log('valuesOfLegendProp:' + JSON.stringify(valuesOfLegendProp, null, 2));
    
    data.Axis.propertyValues = [];
    data.Legend.propertyValues = [];

    if (valuesOfAxisProp.indexOf(null) > -1)
    {
        data.Axis.propertyValues.push('(empty)');  
        data.objectsByPropertyValues['(empty)'] = {};
    }

    for (let i in valuesOfAxisProp)
    {
        const value = valuesOfAxisProp[i];
        if (value != null) {
            data.Axis.propertyValues.push(value);  
            data.objectsByPropertyValues[value] = {};
        }        
    }

    if (valuesOfLegendProp.indexOf(null) > -1)
    {
        data.Legend.propertyValues.push('(empty)');        
    }

    for (let i in valuesOfLegendProp)
    {
        const value = valuesOfLegendProp[i];
        if (value != null)
        {
            data.Legend.propertyValues.push(value);  
        }        
    }
    //console.log(JSON.stringify(data, null, 2))
        
    for (let axisPropValue of data.Axis.propertyValues)
    {       
        data.Axis.objectsByPropertyValues[axisPropValue] = [];
        data.objectsByPropertyValues[axisPropValue] = {};
        for (let legendPropValue of data.Legend.propertyValues)
        {
            data.objectsByPropertyValues[axisPropValue][legendPropValue] = [];
        }        
    }

    for (let objId of data.objects)
    {
        let axisPropValue = await desiteAPI.getPropertyValue(objId, _configCurrent.Properties.Axis.Name, _configCurrent.Properties.Axis.Type);
        let legendPropValue = await desiteAPI.getPropertyValue(objId, _configCurrent.Properties.Legend.Name, _configCurrent.Properties.Legend.Type);
        
        if (axisPropValue == null) axisPropValue = '(empty)';
        if (legendPropValue == null) legendPropValue = '(empty)';
        data.Axis.objectsByPropertyValues[axisPropValue].push(objId);
        data.objectsByPropertyValues[axisPropValue][legendPropValue].push(objId);        
    } 
    return data
}

async function createBookmarkDropdownMenu()
{            
    const bookmarkList = await desiteAPI.getPropertyTypeBookmarkList();
    console.log(JSON.stringify(bookmarkList, null, 2))	
    
    const dropdown = document.getElementById("dropdownBookmark");
    let option = document.createElement("option");
    option.textContent = '';
    option.value = '';
    dropdown.appendChild(option);
    let bookmarkOptId = '';
    
    for (i in bookmarkList)
    {
        const bookmarkObj = bookmarkList[i];
        if (bookmarkObj.Name === ' ') continue;
        option = document.createElement("option");
        option.textContent = bookmarkObj.Name;
        option.value = bookmarkObj.ID;
        dropdown.appendChild(option);
        if (_dataVisualizationConfig.Bookmark == bookmarkObj.Name)
        {            
            bookmarkOptId = bookmarkObj.ID;
        }            
    }
    if (bookmarkOptId.length > 0)
    {
        $('#dropdownBookmark').val(bookmarkOptId);     
    }
    else if (_dataVisualizationConfigFileFound && _dataVisualizationConfig.Bookmark.length > 0 && _dataVisualizationConfig.Bookmark != ' ')
    {
        alert(`Bookmark '${_dataVisualizationConfig.Bookmark}' not found!`);
    } 
    _configCurrent.Bookmark = $("#dropdownBookmark option:selected").text();       
    await onDropdownBookmarkChanged();
}

async function onDropdownBookmarkChanged()
{
    const selectedBookMarkId = $("#dropdownBookmark option:selected").val();
    if (selectedBookMarkId.length > 0)
    {
        _configCurrent.Bookmark = $("#dropdownBookmark option:selected").text();
        const bookmarkObj = await desiteAPI.getPropertyTypeBookmarkByID(selectedBookMarkId);
        const geometryProperties = bookmarkObj.PropertyTypes.geometry;
        console.log(geometryProperties.length);
        console.log(JSON.stringify(geometryProperties, null, 2))
        await createPropertyForAxisDropdownMenu(geometryProperties);        
        await createPropertyForValueDropdownMenu(geometryProperties);        
        await createPropertyForLegendDropdownMenu(geometryProperties);
    }
    else
    {
        clearDropdownMenu("dropdownPropertyForAxis");
        _configCurrent.Properties.Axis = {
            Name: '',
            Type : ''        
        }
        clearDropdownMenu("dropdownPropertyForValue");
        _configCurrent.Properties.Value = {
            Name: '',
            Type : ''        
        }
        clearDropdownMenu("dropdownPropertyForLegend");
        _configCurrent.Properties.Legend = {
            Name: '',
            Type : ''        
        }
        clearDropdownMenu("dropdownPropertyValuesPresentedAs");
        _configCurrent.PropertyValuesPresentedAs = ''; 
    }
}

function clearDropdownMenu(dropdownMenuId){
    const dropdown = document.getElementById(dropdownMenuId);
    $("#" + dropdownMenuId).empty();
    const option = document.createElement("option");
    option.textContent = '';
    option.value = '';
    dropdown.appendChild(option);    
}

async function  onDropdownChartTypeChanged()
{  
    _configCurrent.ChartType = $("#dropdownChartType option:selected").val();
}

async function onDropdownPropertyForAxisChanged()
{
    const propertyNameAndDataType = $("#dropdownPropertyForAxis option:selected").val();
    
    if (propertyNameAndDataType.length > 0)
    {
        _configCurrent.Properties.Axis = {
            Name: propertyNameAndDataType.split("##")[0],
            Type: propertyNameAndDataType.split("##")[1]
        }
    }
    else
    {
        _configCurrent.Properties.Axis = {
            Name: '',
            Type: ''
        }
    }
}

async function onDropdownPropertyForValueChanged()
{
    const propertyNameAndDataType = $("#dropdownPropertyForValue option:selected").val();

    console.log(propertyNameAndDataType)
    if (propertyNameAndDataType.length > 0)
    {
        _configCurrent.Properties.Value = {
            Name: propertyNameAndDataType.split("##")[0],
            Type : propertyNameAndDataType.split("##")[1]
        }
        if (_configCurrent.Properties.Value.Type == 'xs:int' || _configCurrent.Properties.Value.Type == 'xs:long' || _configCurrent.Properties.Value.Type == 'xs:double')
        {
            await createPropertyValuesPresentedAsDropdownMenu(['Count', 'Sum']);
        }
        else
        {
            await createPropertyValuesPresentedAsDropdownMenu(['Count']);
            $('#dropdownPropertyValuesPresentedAs').val('Count');
        }        
    }
    else
    {
        _configCurrent.Properties.Value = {
            Name: '',
            Type : ''
        }
        await createPropertyValuesPresentedAsDropdownMenu([]);        
    }
    await onDropdownPropertyValuesPresentedAsChanged();
}

async function onDropdownPropertyForLegendChanged()
{
    const propertyNameAndDataType = $("#dropdownPropertyForLegend option:selected").val();
    
    if (propertyNameAndDataType.length > 0)
    {
        _configCurrent.Properties.Legend = {
            Name: propertyNameAndDataType.split("##")[0],
            Type : propertyNameAndDataType.split("##")[1]
        }
    }
    else
    {
        _configCurrent.Properties.Legend = {
            Name: '',
            Type : ''
        }
    }        
}

async function  onDropdownPropertyValuesPresentedAsChanged()
{  
    _configCurrent.PropertyValuesPresentedAs = $("#dropdownPropertyValuesPresentedAs option:selected").text();
}

async function onChartTitleChanged()
{
    const element = document.getElementById("tdSectionChartTitle");
    element.innerHTML = $("#inputChartTitle").val();
    _configCurrent.ChartTitle = $("#inputChartTitle").val();
    console.log(' _configCurrent.ChartTitle: ' +  _configCurrent.ChartTitle);
}


async function createPropertyForAxisDropdownMenu(geometryProperties)
{            
    $("#dropdownPropertyForAxis").empty();
    const dropdown = document.getElementById("dropdownPropertyForAxis");
    let option = document.createElement("option");
    option.textContent = '';
    option.value = '';
    dropdown.appendChild(option);

    for (i in geometryProperties)
    {
        const propertyObj = geometryProperties[i];
        option = document.createElement("option");
        option.textContent = propertyObj.name;
        option.value = propertyObj.name + '##' + propertyObj.type;
        dropdown.appendChild(option);
    }
}

async function createPropertyForValueDropdownMenu(geometryProperties)
{            
    $("#dropdownPropertyForValue").empty();
    const dropdown = document.getElementById("dropdownPropertyForValue");
    let option = document.createElement("option");
    option.textContent = '';
    option.value = '';
    dropdown.appendChild(option);

    for (i in geometryProperties)
    {
        const propertyObj = geometryProperties[i];
        option = document.createElement("option");
        option.textContent = propertyObj.name;
        option.value = propertyObj.name + '##' + propertyObj.type;
        dropdown.appendChild(option);
    }
}

async function createPropertyForLegendDropdownMenu(geometryProperties)
{            
    $("#dropdownPropertyForLegend").empty();
    const dropdown = document.getElementById("dropdownPropertyForLegend");
    let option = document.createElement("option");
    option.textContent = '';
    option.value = '';
    dropdown.appendChild(option);

    for (i in geometryProperties)
    {
        const propertyObj = geometryProperties[i];
        option = document.createElement("option");
        option.textContent = propertyObj.name;
        option.value = propertyObj.name + '##' + propertyObj.type;
        dropdown.appendChild(option);
    }
}

async function createPropertyValuesPresentedAsDropdownMenu(optionValues)
{            
    $("#dropdownPropertyValuesPresentedAs").empty();
    const dropdown = document.getElementById("dropdownPropertyValuesPresentedAs");
    let option = document.createElement("option");
    option.textContent = '';
    option.value = '';
    dropdown.appendChild(option);
    
    for (i in optionValues)
    {
        const optionValue = optionValues[i];
        option = document.createElement("option");
        option.textContent = optionValue;
        option.value = optionValue;
        dropdown.appendChild(option);
    }
}

async function setDropdownMenuOptionFromConfig(dropdownId)
{
    const propKey = dropdownId.split('For')[1];
    const propValueFromConfig = _dataVisualizationConfig.Properties[propKey].Name + '##' +  _dataVisualizationConfig.Properties[propKey].Type;
    const element = document.getElementById(dropdownId);
    for (i = 0; i < element.length; i++)
    {        
        if (element.options[i].value == propValueFromConfig)
        {
            $('#' + dropdownId).val(propValueFromConfig);
            _configCurrent.Properties[propKey] = {
                Name: _dataVisualizationConfig.Properties[propKey].Name,
                Type: _dataVisualizationConfig.Properties[propKey].Type
            }
            break;
        }
    }
    console.log('From: setDropdownMenuOptionFromConfig: dropdownId: ' + dropdownId + '\n' + JSON.stringify(_configCurrent.Properties, null, 2)) 
}

async function setDropdownChartTypeFromConfig()
{
    const chartTypeValueFromConfig = _dataVisualizationConfig.ChartType;
    const element = document.getElementById('dropdownChartType');
    for (i = 0; i < element.length; i++)
    {        
        if (element.options[i].value == chartTypeValueFromConfig)
        {
            $('#dropdownChartType').val(chartTypeValueFromConfig); 
            break;
        }
    }
    _configCurrent.ChartType = $("#dropdownChartType option:selected").val();
}

async function setDropdownPropertyValuesPresentedAsFromConfig()
{
    const propForValueOptValue = $("#dropdownPropertyForValue").val();
    if (propForValueOptValue != '')
    {
        const dataType = propForValueOptValue.split('##')[1];
        if (dataType == 'xs:int' || dataType == 'xs:long' || dataType == 'xs:double')
        {
            await createPropertyValuesPresentedAsDropdownMenu(['Count', 'Sum']);            
        }
        else
        {
            await createPropertyValuesPresentedAsDropdownMenu(['Count']);
        }
        const propertyValuesPresentedAsFromConfig = _dataVisualizationConfig.PropertyValuesPresentedAs;
        const element = document.getElementById('dropdownPropertyValuesPresentedAs');
        for (i = 0; i < element.length; i++)
        {        
            if (element.options[i].value == propertyValuesPresentedAsFromConfig)
            {
                $('#dropdownPropertyValuesPresentedAs').val(propertyValuesPresentedAsFromConfig); 
                break;
            }
        }
    }
    _configCurrent.PropertyValuesPresentedAs = $("#dropdownPropertyValuesPresentedAs option:selected").text();
}

async function setChartTitleFromConfig()
{   
    $('#inputChartTitle').val(_dataVisualizationConfig.ChartTitle);
    const element = document.getElementById("tdSectionChartTitle");
    element.innerHTML = $("#inputChartTitle").val();
    _configCurrent.ChartTitle = $("#inputChartTitle").val();
}


async function createSelectPropertyDropdownMenu(dropdownId, geometryProperties)
{            
    $("#"+ dropdownId).empty();
    const dropdown = document.getElementById(dropdownId);
    let option = document.createElement("option");
    option.textContent = '';
    option.value = '';
    dropdown.appendChild(option);

    for (i in geometryProperties)
    {
        const propertyObj = geometryProperties[i];
        option = document.createElement("option");
        option.textContent = propertyObj.name + '##' + propertyObj.type;
        option.value = propertyObj.name + '##' + propertyObj.type;
        dropdown.appendChild(option);
    }
}

async function showChartSettings()
{
    $("#hiddenChartSettingsDiv").hide();
    $("#chartSettingsDiv").show();
}

async function hideChartSettings()
{
    $("#chartSettingsDiv").hide();
    $("#hiddenChartSettingsDiv").show();
}

async function rewriteChartSettingsObject()
{
    await desiteAPI.setAllElementsVisible(false, 'geometry');
    await desiteAPI.setElementsVisible(true, elementIDs);
}

async function onSaveChartSettingsClicked()
{
    console.log(_dataVisualizationConfigFilePath);
    const returns = await desiteAPI.writeFile(_dataVisualizationConfigFilePath,  JSON.stringify(_configCurrent, undefined, 3), "UTF-8");
    console.log(returns);
}

async function checkChartSettings()
{
    _chartSettingsComplete = true;

    if ($("#dropdownBookmark option:selected").text() == '' ||
        $("#dropdownChartType option:selected").val() == '' ||
        $("#dropdownPropertyForAxis option:selected").val() == '' ||
        $("#dropdownPropertyForValue option:selected").val() == '' ||
        $("#dropdownPropertyValuesPresentedAs option:selected").text() == ''||
        $("#inputChartTitle").val() == '')
    {
        _chartSettingsComplete = false;
        alert('Please complete chart settings!');
    }
}
