const UNDEFINED_OPTION = '';

const TriState = 
{
    False: 0,
    True: 1,
    Indeterminate: 2
};

let gTranslator = {};
let gDoorListConfig = {};
let gDoorList = [];
let gDoorMapping = {};
let gSelectedLevelName = UNDEFINED_OPTION;
let gShowOnlyDoors = false;

addEventListener('desiteload', async function(event)
{    
    try
    {
        gTranslator = await createTranslator('/desite.modules/dictionary.json');
        let appLanguage = await desiteAPI.getApplicationLanguage();    
        setCurrentLanguage(gTranslator, appLanguage);
    }
    catch
    {
        console.log("Module translations are not available.");            
    }
    await desiteAPI.selectionChanged.connect(onSelectionChanged);
    gDoorListConfig = await createDoorListConfig('/desite.modules/doorList.module/doorlistConfig.json');
    createDoorPropertyEdit();
    await createLevelDropdownMenu();
    await showElementsBasedOnFilter();
    await createMissingDoorProperties();
    updateDoorList();
});

async function createDoorListConfig(configfilePath)
{
    let config = {};
    let projectPath = await desiteAPI.getProjectDirectory();
    let filePath = projectPath + configfilePath;
    let fileExists = await desiteAPI.checkIfFileExists(filePath);
    if (fileExists)
        config = JSON.parse(await desiteAPI.readTextFileAsString(filePath, "UTF-8"));
    else
        alert(`ERROR: Couldn't find door list config file '${filePath}'`);
    return config;
}

function updateDoorInfo()
{
    $("#tableDoorInfos").empty();

    let hasEntries = false;
    for (i in gDoorList)
    {
        hasEntries = gDoorList[i].IsSelected;
        if (hasEntries)
            break;
    }

    if (!hasEntries)
    {
        $("#message4").show();
        $("#divDoorInfos").hide();
        return;
    }  

    let bgColors = ["#f7f6f6", "#ffffff"];
    let rowIndex = 0;

    let bg = bgColors[(rowIndex++)%2];
    let innerHtml = '<tr>';
    innerHtml += `<td style="background: ${bg}; left: 0; top: 0; z-index: 1; position: sticky">Name</td>`;  
    innerHtml += `<td style="background: ${bg}; top: 0; position: sticky">Level</td>`;    

    let doorProperties = Object.keys(gDoorListConfig.OptionalProperties);
    for (i in doorProperties)
    {
        let doorProperty = gDoorListConfig.OptionalProperties[doorProperties[i]];
        innerHtml += `<td style="background: ${bg}; top: 0; position: sticky">${doorProperty.DisplayName}</td>`; 
    }
    innerHtml += '</tr>';
    
    for (i in gDoorList)
    {
        let door = gDoorList[i];
        if (!door.IsSelected)
            continue;
        let bg = bgColors[(rowIndex++)%2];
        innerHtml += '<tr>';
        innerHtml += `<td style="background: ${bg}; left: 0; position: sticky">${door.Name}</td>`
        innerHtml += `<td style="background: ${bg}; z-index: -1">${door.Level}</td>`
        let doorProperties = Object.keys(gDoorListConfig.OptionalProperties);
        for (i in doorProperties)
        {
            let value = door.OptionalProperties[doorProperties[i]];
            innerHtml += `<td style="background: ${bg}; z-index: -1">${(value != null ? value : 'n/a')}</td>`
        }
        innerHtml += '</tr>';
    }

    $('#tableDoorInfos').append(innerHtml);
    $("#message4").hide();
    $("#divDoorInfos").show();    
}

async function createMissingDoorProperties()
{
    let doorProperties = Object.keys(gDoorListConfig.OptionalProperties);
    for (i in doorProperties)
    {
        let doorProperty = gDoorListConfig.OptionalProperties[doorProperties[i]];
        await desiteAPI.createPropertyType(doorProperty.Name, doorProperty.Type, "", true, 1);
    }
}

async function updateDoorList()
{
    gDoorList = [];
    let elementList = await desiteAPI.getAllElements('geometry');        
    elementList = await desiteAPI.filterByProperty(elementList, gDoorListConfig.Door.Name, gDoorListConfig.Door.Type, gDoorListConfig.Door.PropertyValue, false ,'geometry');
    for (i in elementList)
    {
        let element = elementList[i];
        let doorName = await desiteAPI.getPropertyValue(element, "cpName", "xs:string", false);
        let doorLevel = await desiteAPI.getPropertyValue(element, gDoorListConfig.DoorLevel.Name, gDoorListConfig.DoorLevel.Type, false);
        let isSelected = await desiteAPI.isSelected(element);
        let doorOptionalProperties = {};
        let doorProperties = Object.keys(gDoorListConfig.OptionalProperties);
        for (j in doorProperties)
        {
            let doorProperty = gDoorListConfig.OptionalProperties[doorProperties[j]];
            let value = await desiteAPI.getPropertyValue(element, doorProperty.Name, doorProperty.Type, false);
            doorOptionalProperties[doorProperties[j]] = value;
        }
        let door =
        {
            ID: element,
            Name: doorName,
            Level: doorLevel,
            IsSelected: isSelected,
            OptionalProperties: doorOptionalProperties            
        };
        gDoorMapping[door.ID] = i;
        gDoorList.push(door);        
    }    

    gDoorList.sort(function(doorA, doorB)
    {
        return (doorA.Level > doorB.Level ? 1 : (doorA.Level < doorB.Level ? -1 : 0));
    });

    gDoorMapping = {};
    for (i in gDoorList)
    {
        let door = gDoorList[i];
        gDoorMapping[door.ID] = i;
    }

    updateDoorCheckList();
    updateDoorInfo();
    updateDoorPropertyEdit();   
}

async function setDoorSelected(doorID, isSelected)
{    
    let doorIndex = gDoorMapping[doorID];
    if (doorIndex == undefined) 
        return;
    if (gDoorList[doorIndex].IsSelected == isSelected)
        return;
    gDoorList[doorIndex].IsSelected = isSelected;
    $(`#${doorID}`).prop('checked', isSelected);
    await desiteAPI.setElementsSelected(isSelected, doorID);
}

async function createLevelDropdownMenu()
{            
    let undefinedOption = [UNDEFINED_OPTION];
    let levelNameList = await desiteAPI.getPropertyValues('cpRevitLevel', 'xs:string', 0, 'geometry', false); 
    levelNameList = levelNameList.concat(undefinedOption);
    let sortedLevelNameList = levelNameList.sort();
    let dropdown = document.getElementById('dropdownLevel');
    for (i in sortedLevelNameList)
    {
        let levelName = sortedLevelNameList[i];
        let option = document.createElement('option');
        option.textContent = levelName;
        option.value = levelName;
        dropdown.appendChild(option);
    }
}

function createDoorPropertyEdit()
{
    $("#tableDoorEdit").empty();
    let innerHtml = '<tr>'
    innerHtml += `<td width="50%">Name</td>`;
    innerHtml += `<td><input id="Name" style="width: 295px;" readonly></td>`;
    innerHtml += '</tr>';    
    let doorProperties = Object.keys(gDoorListConfig.OptionalProperties);
    for (i in doorProperties)
    {
        let doorProperty = gDoorListConfig.OptionalProperties[doorProperties[i]];
        if (doorProperty.ReadOnly)
        {
            readOnly = 'readonly';
            disabled = 'disabled="disabled"';
        }else
        {
            readOnly = '';
            disabled = '';
        }
        innerHtml += '<tr>'
        innerHtml += `<td width="50%">${doorProperty.DisplayName}</td>`;
        if (doorProperty.Type == 'xs:boolean')
            innerHtml += `<td><input id="${doorProperties[i]}" onclick="onBooleanEditClicked(this.id)" type="checkbox" ${disabled}/></td>`;
        else if (doorProperty.Type == 'xs:double')
            innerHtml += `<td><input onchange="onStringEditChanged(this.id);" style="width: 295px;" id="${doorProperties[i]}" type="number" min="0" step="0.01" ${readOnly}></td>`;
        else if (doorProperty.Type == 'xs:int')
            innerHtml += `<td><input onchange="onStringEditChanged(this.id);" style="width: 295px;" id="${doorProperties[i]}" type="number" min="0" step="1" ${readOnly}></td>`;
        else
            innerHtml += `<td><input onchange="onStringEditChanged(this.id);" style="width: 295px;" id="${doorProperties[i]}" type="text" ${readOnly}></td>`;
        innerHtml += '</tr>';
    }
    $("#tableDoorEdit").append(innerHtml);    
}

function checkUniqueBoolean(oldValue, newValue)
{
    let result = (oldValue == null ? newValue : (oldValue == newValue ? newValue : TriState.Indeterminate));
    return result;
}

function checkUniqueString(oldValue, newValue)
{
    let result = (oldValue == null ? newValue : (oldValue == newValue ? newValue : ''));
    return result;
}

async function updateDoorPropertyEdit()
{    
    let doorProperties = Object.keys(gDoorListConfig.OptionalProperties);

    let nameList = [];
    let optionalProperties = {};
    for (i in doorProperties)
        optionalProperties[doorProperties[i]] = null;

    for (i in gDoorList)
    {
        let door = gDoorList[i];
        if (!door.IsSelected)
            continue;
        nameList.push(door.Name);
        for (j in doorProperties)
        {
            let doorProperty = gDoorListConfig.OptionalProperties[doorProperties[j]];
            let oldValue = optionalProperties[doorProperties[j]];
            let newValue = door.OptionalProperties[doorProperties[j]];
            if (doorProperty.Type == 'xs:boolean')
                optionalProperties[doorProperties[j]] = checkUniqueBoolean(oldValue, newValue ? TriState.True : TriState.False);
            else
                optionalProperties[doorProperties[j]] = checkUniqueString(oldValue, newValue);
        }
    }
    
    if (nameList.length != 0) 
    {        
        $("#message3").hide();
        $("#tableDoorEdit").show();

        let multipleElementsSelectedStr = 'Multiple elements selected';
        try
        {
            multipleElementsSelectedStr = getTranslatedString(gTranslator, 'multipleElementsSelected', multipleElementsSelectedStr);
        }
        catch
        {
            console.log("Module translations are not available.");
        }

        if (nameList.length == 1) 
            $("#Name").val(nameList[0]).css('font-style', 'normal');
        else
            $("#Name").val(multipleElementsSelectedStr).css('font-style', 'italic');
        
        for (i in doorProperties) 
        {
            let doorProperty = gDoorListConfig.OptionalProperties[doorProperties[i]];
            let value = optionalProperties[doorProperties[i]];
            if (doorProperty.Type == 'xs:boolean') 
            {
                $(`#${doorProperties[i]}`).prop('indeterminate', value == TriState.Indeterminate);
                $(`#${doorProperties[i]}`).prop('checked', value == TriState.True);
            }
            else 
            {
                $(`#${doorProperties[i]}`).val(value);
            }
        }        
    } 
    else 
    {
        $("#message3").show();
        $("#tableDoorEdit").hide();
    }
}

async function showElementsBasedOnFilter()
{        
    let elementList = await desiteAPI.getAllElements('geometry');
    if (gSelectedLevelName != UNDEFINED_OPTION)
    {
        let levelNameFilter = `"${gSelectedLevelName}"`;
        elementList = await desiteAPI.filterByProperty(elementList, 'cpRevitLevel', 'xs:string', levelNameFilter, true ,'geometry');
    }    
    if (gShowOnlyDoors)
    {        
        elementList = await desiteAPI.filterByProperty(elementList, 'cpRevitCategory', 'xs:string', 'Doors', true ,'geometry');
    }
    await desiteAPI.setAllElementsVisible(false, 'geometry')
    await desiteAPI.setElementsVisible(true, elementList);
}

function updateDoorCheckList()
{
    $("#tableDoorList").empty();
    for (i in gDoorList)
    {
        let door = gDoorList[i];
        if (gSelectedLevelName != UNDEFINED_OPTION && door.Level != gSelectedLevelName)
            continue;
        let checked = (door.IsSelected ? 'checked' : '');
        $('#tableDoorList').append(`<tr><td><input type="checkbox" id="${door.ID}" onclick="onDoorSelectionChanged(this.id)" ${checked}/></td><td>${door.Name} - ${door.Level}</td></tr>`);        
    } 

    if ($('#tableDoorList tr').length != 0)
    {
        $("#message1").show();
        $("#message2").hide();        
        $("#divDoorList").show();
    }
    else
    {
        $("#message1").hide();
        $("#message2").show();
        $("#divDoorList").hide();
    }     
}

async function onBooleanEditClicked(id)
{
    let isChecked = $(`#${id}`).is(":checked");
    let doorProperty = gDoorListConfig.OptionalProperties[id];
    for (i in gDoorList)
    {
        let door = gDoorList[i];
        if (!door.IsSelected)
            continue;
        desiteAPI.setPropertyValue(door.ID, doorProperty.Name, doorProperty.Type, isChecked);
        door.OptionalProperties[id] = isChecked;
    }
    updateDoorInfo();
    updateDoorPropertyEdit();   
}

async function onStringEditChanged(id)
{
    let value = $(`#${id}`).val();
    let doorProperty = gDoorListConfig.OptionalProperties[id];
    for (i in gDoorList)
    {
        let door = gDoorList[i];
        if (!door.IsSelected)
            continue;
        desiteAPI.setPropertyValue(door.ID, doorProperty.Name, doorProperty.Type, value);
        door.OptionalProperties[id] = value;
    }
    updateDoorInfo();
    updateDoorPropertyEdit();   
}

async function onLevelDropdownMenuChanged() 
{
    let dropdown = document.getElementById('dropdownLevel');
    let selectedIndex = dropdown.selectedIndex;
    gSelectedLevelName = dropdown.options[selectedIndex].value;
    showElementsBasedOnFilter();    
    updateDoorCheckList();
}

async function onDoorCheckboxClicked()
{
    gShowOnlyDoors = $("#checkboxDoor").is(":checked");
    showElementsBasedOnFilter();
}

async function onDoorSelectionChanged(id)
{
    let isChecked = $(`#${id}`).is(":checked");
    await setDoorSelected(id, isChecked);
}

async function onSelectionChanged() 
{
    let selectElementList = await desiteAPI.getSelectedElements('geometry');
    let selectElementSet = new Set(selectElementList);
    for (i in gDoorList)
    {
        let door = gDoorList[i];
        let isSelected = selectElementSet.has(door.ID);
        await setDoorSelected(door.ID, isSelected);
    }
    updateDoorInfo();
    updateDoorPropertyEdit();   
}

async function onPrintClicked()
{
    await desiteMD.setPageMargins(20,20,20,20);
    await desiteMD.setOrientation(1);   
    await desiteMD.setPaperSize(0);
    let textBlocks =
    {
        content: $('#tableDoorInfos').html()
    };
    let filePath = await desiteAPI.getProjectDirectory() + '/desite.modules/doorList.module/printTemplate.html';
    await desiteMD.createPdfByTemplate(filePath,'', textBlocks);
}