const UNDEFINED_OPTION = '';

const TriState = 
{
    False: 0,
    True: 1,
    Indeterminate: 2
};

let gTranslator = {};
let gRoombookConfig = {};
let gRoomList = [];
let gRoomMapping = {};
let gSelectedLevelName = UNDEFINED_OPTION;
let gSelectedRoomTypeName = UNDEFINED_OPTION;

addEventListener('desiteload', async function(event)
{            
    try
    {
        gTranslator = await createTranslator('/desite.modules/dictionary.json');
        let appLanguage = await desiteAPI.getApplicationLanguage();
        setCurrentLanguage(gTranslator, appLanguage);    
    }
    catch
    {
        console.log("Module translations are not available.");
    }
    await desiteAPI.selectionChanged.connect(onSelectionChanged);
    gRoombookConfig = await createRoombookConfig('/desite.modules/roombook.module/roombookConfig.json');
    createRoomPropertyEdit();    
    await createLevelDropdownMenu();
    await createRoomTypeDropdownMenu();
    await createMissingRoomProperties();        
    updateRoomList();
    showElementsBasedOnFilter();    
});            

async function createRoombookConfig(configfilePath)
{
    let config = {};
    let projectPath = await desiteAPI.getProjectDirectory();
    let filePath = projectPath + configfilePath;
    let fileExists = await desiteAPI.checkIfFileExists(filePath);
    if (fileExists)
        config = JSON.parse(await desiteAPI.readTextFileAsString(filePath, "UTF-8"));
    else
        alert(`ERROR: Couldn't find roombook config file '${filePath}'`);
    return config;
}

async function createMissingRoomProperties()
{
    let roomProperties = Object.keys(gRoombookConfig.OptionalProperties);
    for (i in roomProperties)
    {
        let roomProperty = gRoombookConfig.OptionalProperties[roomProperties[i]];
        await desiteAPI.createPropertyType(roomProperty.Name, roomProperty.Type, "", true, 1);
    }
}

async function createLevelDropdownMenu()
{            
    let undefinedOption = [UNDEFINED_OPTION];
    let roomLevel = gRoombookConfig.RoomLevel;
    let roomLevelNameList = await desiteAPI.getPropertyValues(roomLevel.Name, roomLevel.Type, 0, "geometry", false);
    roomLevelNameList = roomLevelNameList.concat(undefinedOption);
    let sortedRoomLevelNameList = roomLevelNameList.sort();
    let dropdown = document.getElementById("dropdownLevel");
    for (i in sortedRoomLevelNameList)
    {
        let levelName = sortedRoomLevelNameList[i];
        let option = document.createElement("option");
        option.textContent = levelName;
        option.value = levelName;
        dropdown.appendChild(option);
    }
}

async function createRoomTypeDropdownMenu()
{            
    let undefinedOption = [UNDEFINED_OPTION];
    let roomType = gRoombookConfig.RoomType;
    let roomTypeNameList = await desiteAPI.getPropertyValues(roomType.Name, roomType.Type, 0, "geometry", false);
    roomTypeNameList = roomTypeNameList.concat(undefinedOption);
    let sortedRoomTypeNameList = roomTypeNameList.sort();
    let dropdown = document.getElementById("dropdownRoomType");
    for (i in sortedRoomTypeNameList)
    {
        let roomTypeName = sortedRoomTypeNameList[i];
        let option = document.createElement("option");
        option.textContent = roomTypeName;
        option.value = roomTypeName;
        dropdown.appendChild(option);
    }
}

function createRoomPropertyEdit()
{
    $("#tableRoomEdit").empty();
    let innerHtml = '<tr>'
    innerHtml += `<td width="50%">Name</td>`;
    innerHtml += `<td><input id="Name" style="width: 295px;" readonly></td>`;
    innerHtml += '</tr>';    
    let roomProperties = Object.keys(gRoombookConfig.OptionalProperties);
    for (i in roomProperties)
    {
        let roomProperty = gRoombookConfig.OptionalProperties[roomProperties[i]];
        if(roomProperty.ReadOnly)
        {
            readOnly = 'readonly';
            disabled = 'disabled="disabled"';
        }
        else
        {
            readOnly = '';
            disabled = '';
        }
        
        innerHtml += '<tr>'
        innerHtml += `<td width="50%">${roomProperty.DisplayName}</td>`;
        if (roomProperty.Type == 'xs:boolean')
            innerHtml += `<td><input id="${roomProperties[i]}" onclick="onBooleanEditClicked(this.id)" type="checkbox" ${disabled}/></td>`;
        else if (roomProperty.Type == 'xs:double')
            innerHtml += `<td><input onchange="onStringEditChanged(this.id);" style="width: 295px;" id="${roomProperties[i]}" type="number" min="0" step="0.01" ${readOnly}></td>`;
        else if (roomProperty.Type == 'xs:int')
            innerHtml += `<td><input onchange="onStringEditChanged(this.id);" style="width: 295px;" id="${roomProperties[i]}" type="number" min="0" step="1" ${readOnly}></td>`;
        else
            innerHtml += `<td><input onchange="onStringEditChanged(this.id);" style="width: 295px;" id="${roomProperties[i]}" type="text" ${readOnly}></td>`;
        innerHtml += '</tr>';
    }
    $("#tableRoomEdit").append(innerHtml);    
}

function checkUniqueBoolean(oldValue, newValue)
{
    let result = (oldValue == null ? newValue : (oldValue == newValue ? newValue : TriState.Indeterminate));
    return result;
}

function checkUniqueString(oldValue, newValue)
{
    let result = (oldValue == null ? newValue : (oldValue == newValue ? newValue : ''));
    return result;
}

async function updateRoomPropertyEdit()
{    
    let roomProperties = Object.keys(gRoombookConfig.OptionalProperties);

    let nameList = [];
    let optionalProperties = {};
    for (i in roomProperties)
        optionalProperties[roomProperties[i]] = null;

    for (i in gRoomList)
    {
        let room = gRoomList[i];
        if (!room.IsSelected)
            continue;
        nameList.push(room.Name);
        for (j in roomProperties)
        {
            let roomProperty = gRoombookConfig.OptionalProperties[roomProperties[j]];
            let oldValue = optionalProperties[roomProperties[j]];
            let newValue = room.OptionalProperties[roomProperties[j]];
            if (roomProperty.Type == 'xs:boolean')
                optionalProperties[roomProperties[j]] = checkUniqueBoolean(oldValue, newValue ? TriState.True : TriState.False);
            else
                optionalProperties[roomProperties[j]] = checkUniqueString(oldValue, newValue);
        }
    }

    if (nameList.length != 0) 
    {
        $("#message3").hide();
        $("#tableRoomEdit").show();

        let multipleElementsSelectedStr = 'Multiple elements selected';
        try
        {
            multipleElementsSelectedStr = getTranslatedString(gTranslator, 'multipleElementsSelected', multipleElementsSelectedStr);
        }
        catch
        {
            console.log("Module translations are not available.");
        }

        if (nameList.length == 1) 
            $("#Name").val(nameList[0]).css('font-style', 'normal');
        else
            $("#Name").val(multipleElementsSelectedStr).css('font-style', 'italic');

        for (i in roomProperties)
        {
            let roomProperty = gRoombookConfig.OptionalProperties[roomProperties[i]];
            let value = optionalProperties[roomProperties[i]];
            if (roomProperty.Type == 'xs:boolean')
            {
                $(`#${roomProperties[i]}`).prop('indeterminate', value == TriState.Indeterminate);
                $(`#${roomProperties[i]}`).prop('checked', value == TriState.True);
            }
            else
            {
                $(`#${roomProperties[i]}`).val(value);
            }
        }
    } 
    else 
    {
        $("#message3").show();
        $("#tableRoomEdit").hide();
    }    
}

function updateRoomInfo()
{
    $("#tableRoomInfos").empty();
    let hasEntries = false;
    for (i in gRoomList)
    {
        hasEntries = gRoomList[i].IsSelected;
        if (hasEntries)
            break;
    }
    if (!hasEntries)
    {
        $("#message4").show();
        $("#divRoomInfos").hide();
        return;
    }  
    
    let bgColors = ["#f7f6f6", "#ffffff"];
    let rowIndex = 0;

    let bg = bgColors[(rowIndex++)%2];
    let innerHtml = '<tr>';
    innerHtml += `<td style="background: ${bg}; z-index: 1; left: 0; top: 0; position: sticky">Name</td>`;  
    innerHtml += `<td style="background: ${bg}; top: 0; position: sticky">Level</td>`;                      
    innerHtml += `<td style="background: ${bg}; top: 0; position: sticky">Type</td>`;                       

    let roomProperties = Object.keys(gRoombookConfig.OptionalProperties);
    for (i in roomProperties)
    {
        let roomProperty = gRoombookConfig.OptionalProperties[roomProperties[i]];
        innerHtml += `<td style="background: ${bg}; top:0; position: sticky">${roomProperty.DisplayName}</td>`;  
    }
    innerHtml += '</tr>';
    
    for (i in gRoomList)
    {
        let room = gRoomList[i];
        if (!room.IsSelected)
            continue;
        let bg = bgColors[(rowIndex++)%2];
        innerHtml += '<tr>';
        innerHtml += `<td style="background: ${bg}; left: 0; position: sticky">${room.Name}</td>`
        innerHtml += `<td style="background: ${bg}; z-index: -1">${room.Level}</td>`
        innerHtml += `<td style="background: ${bg}; z-index: -1">${room.Type}</td>`
        let roomProperties = Object.keys(gRoombookConfig.OptionalProperties);
        for (i in roomProperties)
        {
            let value = room.OptionalProperties[roomProperties[i]];
            innerHtml += `<td style="background: ${bg}; z-index: -1">${(value != null ? value : 'n/a')}</td>`
        }
        innerHtml += '</tr>';
    }    

    $('#tableRoomInfos').append(innerHtml);
    $("#message4").hide();
    $("#divRoomInfos").show();  
}

async function updateRoomList()
{    
    gRoomList = [];
    let elementList = await desiteAPI.getAllElements('geometry');        
    elementList = await desiteAPI.filterByProperty(elementList, gRoombookConfig.Room.Name, gRoombookConfig.Room.Type, gRoombookConfig.Room.PropertyValue, false ,'geometry');
    for (i in elementList)
    {
        let element = elementList[i];
        let roomName = await desiteAPI.getPropertyValue(element, "cpName", "xs:string", false);
        let roomLevel = await desiteAPI.getPropertyValue(element, gRoombookConfig.RoomLevel.Name, gRoombookConfig.RoomLevel.Type, false);
        let roomType = await desiteAPI.getPropertyValue(element, gRoombookConfig.RoomType.Name, gRoombookConfig.RoomType.Type, false);
        let isSelected = await desiteAPI.isSelected(element);
        let roomOptionalProperties = {};
        let roomProperties = Object.keys(gRoombookConfig.OptionalProperties);
        for (j in roomProperties)
        {
            let roomProperty = gRoombookConfig.OptionalProperties[roomProperties[j]];
            let value = await desiteAPI.getPropertyValue(element, roomProperty.Name, roomProperty.Type, false);
            roomOptionalProperties[roomProperties[j]] = value;
        }
        let room =
        {
            ID: element,
            Name: roomName,
            Level: roomLevel,
            Type: roomType,
            IsSelected: isSelected,
            OptionalProperties: roomOptionalProperties            
        };        
        gRoomList.push(room);
    }    

    gRoomList.sort(function(roomA, roomB)
    {
        return (roomA.Level > roomB.Level ? 1 : (roomA.Level < roomB.Level ? -1 : 0));
    });

    gRoomMapping = {};
    for (i in gRoomList)
    {
        let room = gRoomList[i];
        gRoomMapping[room.ID] = i;
    }

    updateRoomCheckList();
    updateRoomInfo();    
    updateRoomPropertyEdit();
}

async function setRoomSelected(roomID, isSelected)
{    
    let roomIndex = gRoomMapping[roomID];
    if(roomIndex == undefined)
        return;
    if(gRoomList[roomIndex].IsSelected == isSelected)
        return;
    gRoomList[roomIndex].IsSelected = isSelected;
    $(`#${roomID}`).prop('checked', isSelected);
    await desiteAPI.setElementsSelected(isSelected, roomID);
}

async function onRoomSelectionChanged(id)
{
    let isChecked = $(`#${id}`).is(":checked");
    await setRoomSelected(id, isChecked);
    updateRoomInfo();
    updateRoomPropertyEdit();
}

async function onSelectionChanged()
{
    let selectElementList = await desiteAPI.getSelectedElements('geometry');
    let selectElementSet = new Set(selectElementList);
    for (i in gRoomList)
    {
        let room = gRoomList[i];
        let isSelected = selectElementSet.has(room.ID);
        await setRoomSelected(room.ID, isSelected);
    }
    updateRoomInfo();
    updateRoomPropertyEdit();
}

async function onBooleanEditClicked(id)
{
    let isChecked = $(`#${id}`).is(":checked");
    let roomProperty = gRoombookConfig.OptionalProperties[id];
    for (i in gRoomList)
    {
        let room = gRoomList[i];
        if (!room.IsSelected)
            continue;
        desiteAPI.setPropertyValue(room.ID, roomProperty.Name, roomProperty.Type, isChecked);
        room.OptionalProperties[id] = isChecked;
    }
    updateRoomInfo();
    updateRoomPropertyEdit();   
}

async function onStringEditChanged(id)
{
    let value = $(`#${id}`).val();
    let roomProperty = gRoombookConfig.OptionalProperties[id];
    for (i in gRoomList)
    {
        let room = gRoomList[i];
        if (!room.IsSelected)
            continue;
        desiteAPI.setPropertyValue(room.ID, roomProperty.Name, roomProperty.Type, value);
        room.OptionalProperties[id] = value;
    }
    updateRoomInfo();
    updateRoomPropertyEdit();   
}

function updateRoomCheckList()
{
    $("#tableRoomList").empty();
    for (i in gRoomList)
    {
        let room = gRoomList[i];
        if(gSelectedLevelName != UNDEFINED_OPTION && room.Level != gSelectedLevelName)
            continue;
        if(gSelectedRoomTypeName != UNDEFINED_OPTION && room.Type != gSelectedRoomTypeName)
            continue;
        let checked = (room.IsSelected ? 'checked' : '');        
        $('#tableRoomList').append(`<tr><td><input type="checkbox" id="${room.ID}" onclick="onRoomSelectionChanged(this.id)" ${checked}/></td><td>${room.Name} - ${room.Level} - ${room.Type}</td></tr>`);        
    } 
    
    if ($('#tableRoomList tr').length != 0)
    {
        $("#message2").hide();
        $("#message1").show();
        $("#divRoomList").show();
    }
    else
    {
        $("#message1").hide();
        $("#message2").show();
        $("#divRoomList").hide();
    }
}

async function showElementsBasedOnFilter()
{
    let elementList = await desiteAPI.getAllElements('geometry');        
    elementList = await desiteAPI.filterByProperty(elementList, gRoombookConfig.Room.Name, gRoombookConfig.Room.Type, gRoombookConfig.Room.PropertyValue, false ,'geometry');
    if (gSelectedLevelName != UNDEFINED_OPTION)
    {
        let levelNameFilter = `"${gSelectedLevelName}"`;
        elementList = await desiteAPI.filterByProperty(elementList, gRoombookConfig.RoomLevel.Name, gRoombookConfig.RoomLevel.Type, levelNameFilter, false ,'geometry');
    }
    if (gSelectedRoomTypeName != UNDEFINED_OPTION)
    {
        let roomTypeNameFilter = `"${gSelectedRoomTypeName}"`;
        elementList = await desiteAPI.filterByProperty(elementList, gRoombookConfig.RoomType.Name, gRoombookConfig.RoomType.Type, roomTypeNameFilter, false ,'geometry');
    }
    await desiteAPI.setAllElementsVisible(false, 'geometry')
    await desiteAPI.setElementsVisible(true, elementList);
}

async function onDropdownLevelChanged()
{
    let selectedOption = document.getElementById('dropdownLevel');
    let selectedIndex = selectedOption.selectedIndex;
    gSelectedLevelName = selectedOption.options[selectedIndex].value;
    showElementsBasedOnFilter();   
    updateRoomCheckList();
}

async function onDropdownRoomTypeChanged()
{
    let selectedOption = document.getElementById('dropdownRoomType');
    let selectedIndex = selectedOption.selectedIndex;
    gSelectedRoomTypeName = selectedOption.options[selectedIndex].value;
    showElementsBasedOnFilter();   
    updateRoomCheckList();
}

async function onPrintClicked()
{
    await desiteMD.setPageMargins(20,20,20,20);
    await desiteMD.setOrientation(1);   
    await desiteMD.setPaperSize(0);
    let textBlocks =
    {
        content: $('#tableRoomInfos').html()
    };
    let filePath = await desiteAPI.getProjectDirectory() + '/desite.modules/roombook.module/printTemplate.html';
    await desiteMD.createPdfByTemplate(filePath,'', textBlocks);
}