const UNDEFINED_OPTION = '';

const StatusType = 
{
    None: 0,
    All: 1,
    Overdue: 2,
    OnTime: 3
};

let gTranslator = {};
let gSelectedLevelName = UNDEFINED_OPTION;
let gSelectedCategoryName = UNDEFINED_OPTION;
let gShowOnlyStatus = false;
let gCurrentStatus = StatusType.None;
let gOverdueMaterialName = "Overdue";
let gOnTimeMaterialName = "On time";

addEventListener('desiteload', async function(event)
{
    try    
    {
        gTranslator = await createTranslator('/desite.modules/dictionary.json');
        let appLanguage = await desiteAPI.getApplicationLanguage();
        setCurrentLanguage(gTranslator, appLanguage);    
        setDatepickerLanguage(appLanguage);
    }
    catch
    {
        console.log("Module translations are not available.");
    }    
    await desiteAPI.selectionChanged.connect(onSelectionChanged);
    await createMaterials();
    await createLevelDropdownMenu();
    await createCategoryDropdownMenu();    
    createStatusDropdownMenu();        
    await desiteAPI.resetMaterials(true);
    await showElementsBasedOnFilter();           
    await updateSelectedElementInfo();        
});

function onSelectionChanged()
{
    updateSelectedElementInfo();
}

/*
Creates dropdown menu options for available categories in the model, sorted in ascending order
*/
async function createCategoryDropdownMenu()
{
    let undefinedOption = [UNDEFINED_OPTION];
    let categoryNameList = await desiteAPI.getPropertyValues("cpRevitCategory", "xs:string", 0 ,"geometry", false);
    categoryNameList = categoryNameList.concat(undefinedOption);
    let sortedCategoryNameList = categoryNameList.sort();
    let dropdown = document.getElementById("dropdownCategory");
    for (i in sortedCategoryNameList)
    {
        let categoryName = sortedCategoryNameList[i];
        let option = document.createElement("option");
        option.textContent = categoryName;
        option.value =  categoryName;
        dropdown.appendChild(option);
    }
}

/*
Creates dropdown menu options for available levels in the model, sorted in ascending order.
*/
async function createLevelDropdownMenu()
{            
    let undefinedOption = [UNDEFINED_OPTION];
    let levelNameList = await desiteAPI.getPropertyValues("cpRevitLevel", "xs:string", 0, "geometry", false);
    levelNameList = levelNameList.concat(undefinedOption);
    let sortedLevelNameList = levelNameList.sort();
    let dropdown = document.getElementById("dropdownLevel");
    for (i in sortedLevelNameList)
    {
        let levelName = sortedLevelNameList[i];
        let option = document.createElement("option");
        option.textContent = levelName;
        option.value = levelName;
        dropdown.appendChild(option);
    }
}

function createStatusDropdownMenu()
{            
    let options = [UNDEFINED_OPTION, "All", "Overdue", "On Time"];
    let optionValues = [StatusType.None, StatusType.All, StatusType.Overdue, StatusType.OnTime];
    let dropdown = document.getElementById("dropdownStatus");
    for (i in options)
    {
        let option = document.createElement("option");
        option.textContent = options[i];
        option.value = optionValues[i];
        dropdown.appendChild(option);
    }
}

function doubleDigitString(number)
{
    return new String(number < 10 ? ('0' + number) : number);
}

function convertUTCDateToDatepickerFormat(dateUTC)
{
    let date = new Date(dateUTC);
    let day = doubleDigitString(date.getDate());
    let month = doubleDigitString(date.getMonth() + 1); // months are zero-based, so we have to increment by one.
    let year = date.getFullYear();
    return `${day}.${month}.${year}`;
}

function convertDatepickerFormatToUTCDate(value, time)
{
    let date = new Date();
    let splitValue = value.split('.');
    if(splitValue.length == 3)
    {
        let day = splitValue[0];
        let month = splitValue[1];
        let year = splitValue[2];
        date = new Date(`${year}-${month}-${day}${time}`);
    }
    else
    {
        alert("ERROR: Invalid date picker value format.");
    }
    return date;
}

/*
Updates the GUI for the first linked activity of the currently selected element.
Currently only a single element and it's first linked activity are supported.
*/
async function updateSelectedElementInfo()
{
    let name = '';
    let linkedTasksCount = '';
    let activityName = '';
    let startDate = '';
    let finishDate = '';
    let actualFinishDate = '';

    $('#inputDateStart').prop('disabled', true);
    $('#inputDateFinish').prop('disabled', true);
    $('#inputDateActualFinish').prop('disabled', true);

    let selectedElements = await desiteAPI.getSelectedElements();
    if (selectedElements.length == 0)
    {
        $("#message").show();
        $("#tableScheduleEdit").hide();
    }
    else
    {
        $("#message").hide();
        $("#tableScheduleEdit").show();

        let noActivitiesWereFoundStr = 'No activities were found that are linked to the selected element!';
        let numberOfSelectedElementsStr = 'Number of selected elements';
        let infoStr = 'Currently only a single element and its first linked activity are supported for editing!';
        try
        {
            noActivitiesWereFoundStr = getTranslatedString(gTranslator, "noActivitiesWereFound", noActivitiesWereFoundStr);
            numberOfSelectedElementsStr = getTranslatedString(gTranslator, 'numberOfSelectedElements', numberOfSelectedElementsStr);
            infoStr = getTranslatedString(gTranslator, 'singleElementAndFirstActivityIsSupported', infoStr);
        }
        catch
        {
            console.log("Module translations are not available.");
        }

        if (selectedElements.length == 1)
        {
            let firstSelectedElement = selectedElements[0];
            name = await desiteAPI.getPropertyValue(firstSelectedElement, 'cpName', 'xs:string');
            let linkedActivities = await desiteAPI.getLinkedObjects(firstSelectedElement, 'activities', 'geometry');
            if (linkedActivities.length > 0)
            {
                let firstLinkedActivity = linkedActivities[0];
                linkedTasksCount = await desiteAPI.getPropertyValue(firstSelectedElement, 'cpCountLinkedTasks', 'xs:int');                
                activityName = await desiteAPI.getPropertyValue(firstLinkedActivity, 'cpName', 'xs:string');
                let startDateUTC = await desiteAPI.getPropertyValue(firstLinkedActivity, 'Start', 'xs:dateTime');
                let finishDateUTC = await desiteAPI.getPropertyValue(firstLinkedActivity, 'Finish', 'xs:dateTime');
                let actualFinishDateUTC = await desiteAPI.getPropertyValue(firstLinkedActivity, 'ActualFinish', 'xs:dateTime');
                startDate = (startDateUTC ? convertUTCDateToDatepickerFormat(startDateUTC) : '');
                finishDate = (finishDateUTC ? convertUTCDateToDatepickerFormat(finishDateUTC) : '');
                actualFinishDate = (actualFinishDateUTC ? convertUTCDateToDatepickerFormat(actualFinishDateUTC) : '');
                $('#inputDateStart').prop('disabled', false);
                $('#inputDateFinish').prop('disabled', false);
                $('#inputDateActualFinish').prop('disabled', false);
            }
            else
            {
                linkedTasksCount = 0;                
                alert(`${noActivitiesWereFoundStr}`);
            }
        }
        else
        {
            alert(`${numberOfSelectedElementsStr}: ${selectedElements.length}\n${infoStr}`);            
        }
    }    

    $('#elementName').val(name);
    $('#activityName').val(activityName);
    $('#linkedTasksCount').val(linkedTasksCount);
    $('#inputDateStart').val(startDate);
    $('#inputDateFinish').val(finishDate);
    $('#inputDateActualFinish').val(actualFinishDate);
}

const DEFAULT_DATE_FORMAT = 'dd.mm.yy';
const DEFAULT_START_DAY = 1; // To convert from US format where sunday is the first day of the week to European format where monday is the first day of the week.
const DEFAULT_SHOW_WEEK = true;
const DEFAULT_SHOW_OTHER_MONTHS = true;

function datepickerOptionsGerman()
{
    let dateOptions =
    {
        dateFormat: DEFAULT_DATE_FORMAT,
        showWeek: DEFAULT_SHOW_WEEK,
        showOtherMonths: DEFAULT_SHOW_OTHER_MONTHS,
        firstDay: DEFAULT_START_DAY, 
        weekHeader: 'KW',
        monthNames: ["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"],
        monthNamesShort: ["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"],
        dayNames: ["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"],
        dayNamesShort: ["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"],
        dayNamesMin:["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]
    };
    return dateOptions;
}

function datepickerOptionsEnglish()
{
    let dateOptions =
    {
        dateFormat: DEFAULT_DATE_FORMAT,
        showWeek: DEFAULT_SHOW_WEEK,
        showOtherMonths: DEFAULT_SHOW_OTHER_MONTHS,
        firstDay: DEFAULT_START_DAY
    };
    return dateOptions;
}

function setDatepickerLanguage(currentLanguage)
{           
    let htmlElements = "#inputDateStart, #inputDateFinish, #inputDateActualFinish";        
    let datepickerOptions = (currentLanguage == "de" ? datepickerOptionsGerman() : datepickerOptionsEnglish())    
    $(htmlElements).datepicker("destroy");
    $(htmlElements).datepicker(datepickerOptions);
}

async function showElementsBasedOnFilter()
{        
    let elementList = await desiteAPI.getAllElements('geometry');
    if (gSelectedLevelName != UNDEFINED_OPTION)
    {
        let levelNameFilter = `"${gSelectedLevelName}"`;
        elementList = await desiteAPI.filterByProperty(elementList, 'cpRevitLevel', 'xs:string', levelNameFilter, true ,'geometry');
    }    
    if (gSelectedCategoryName != UNDEFINED_OPTION)
    {        
        let categoryNameFilter = `"${gSelectedCategoryName}"`;    
        elementList = await desiteAPI.filterByProperty(elementList, 'cpRevitCategory', 'xs:string', categoryNameFilter, true ,'geometry');
    }
    if (gShowOnlyStatus && gCurrentStatus != StatusType.None)
    {
        let materialNameList = [];
        if (gCurrentStatus == StatusType.OnTime || gCurrentStatus == StatusType.All)
            materialNameList.push(gOnTimeMaterialName);
        if (gCurrentStatus == StatusType.Overdue || gCurrentStatus == StatusType.All)
            materialNameList.push(gOverdueMaterialName);
        let materialNameFilter = ''
        for (i in materialNameList)        
            materialNameFilter += ((i == (materialNameList.length - 1)) ? `"${materialNameList[i]}"` : `"${materialNameList[i]}" `);
        elementList = await desiteAPI.filterByProperty(elementList, 'cpMaterialName', 'xs:string', materialNameFilter, false ,'geometry');
    }
    await desiteAPI.setAllElementsVisible(false, 'geometry')
    await desiteAPI.setElementsVisible(true, elementList);
}

async function onLevelDropdownMenuChanged() 
{
    let dropdown = document.getElementById('dropdownLevel');
    let selectedIndex = dropdown.selectedIndex;
    gSelectedLevelName = dropdown.options[selectedIndex].value;
    showElementsBasedOnFilter();    
}

async function onCategoryDropdownMenuChanged()
{
    let dropdown = document.getElementById('dropdownCategory');
    let selectedIndex = dropdown.selectedIndex;
    gSelectedCategoryName = dropdown.options[selectedIndex].value;
    showElementsBasedOnFilter();
}

async function createAndActivateColorSchema() 
{
    let colorSchemaName = "Schedule Status";
    let colorSchemaOptions =
    {
        legend: true
    };
    await desiteAPI.createColorSchemaFromCurrentMaterials(colorSchemaName, true, colorSchemaOptions);
    await desiteAPI.activateColorSchemaByName(colorSchemaName);
}

async function createMaterials()
{
    await desiteAPI.createMaterial('<material ID="' + gOverdueMaterialName + '" name="' + gOverdueMaterialName + '">\
    <diff r="250" b="0" g="0" a="1"/>\
    <amb r="250" b="0" g="0" a="1"/>\
    <spec r="255" b="255" g="255" a="1"/>\
    <trans v="0"/>\
    <shin v="0.5"/>\
    </material>');
    
    await desiteAPI.createMaterial('<material ID="' + gOnTimeMaterialName + '" name="' + gOnTimeMaterialName + '">\
    <diff r="0" b="0" g="130" a="1"/>\
    <amb r="0" b="0" g="130" a="1"/>\
    <spec r="255" b="255" g="255" a="1"/>\
    <trans v="0"/>\
    <shin v="0.5"/>\
    </material>');
}

async function createElementListForTimings()
{
    let timings = 
    {
        onTimeElementList: [],
        overdueElementList: []
    };
    let activityElementList = await desiteAPI.getAllElements('activities');     
    activityElementList = await desiteAPI.filterByProperty(activityElementList, 'Finish', 'xs:dateTime', '*', false ,'activities');
    activityElementList = await desiteAPI.filterByProperty(activityElementList, 'ActualFinish', 'xs:dateTime', '*', false ,'activities');
    for (i in activityElementList)
    {
        let activityElement = activityElementList[i];
        let finishDate = new Date(await desiteAPI.getPropertyValue(activityElement, "Finish", "xs:dateTime"));
        let actualFinishDate = new Date(await desiteAPI.getPropertyValue(activityElement, "ActualFinish", "xs:dateTime"));
        let linkedElementList = await desiteAPI.getLinkedObjects(activityElement, "geometry", "activities");
        for (j in linkedElementList) 
        {
            let linkedElement = linkedElementList[j];
            if (finishDate >= actualFinishDate)
                timings.onTimeElementList.push(linkedElement);
            else if (finishDate < actualFinishDate)
                timings.overdueElementList.push(linkedElement);
        }
    }
    return timings;
}

async function setCurrentStatus(status)
{
    gCurrentStatus = status;
    if (gShowOnlyStatus)
        await showElementsBasedOnFilter();
}

async function applyStatusAll()
{
    let timings = await createElementListForTimings();
    await desiteAPI.resetMaterials(true);
    await desiteAPI.setMaterialToObjects(gOnTimeMaterialName, timings.onTimeElementList);
    await desiteAPI.setMaterialToObjects(gOverdueMaterialName, timings.overdueElementList);
    await createAndActivateColorSchema();
    await setCurrentStatus(StatusType.All);
}

async function applyStatusOnTime()
{
    let timings = await createElementListForTimings();
    await desiteAPI.resetMaterials(true);
    await desiteAPI.setMaterialToObjects(gOnTimeMaterialName, timings.onTimeElementList);
    await createAndActivateColorSchema();
    await setCurrentStatus(StatusType.OnTime);
}

async function applyStatusOverdue()
{
    let timings = await createElementListForTimings();
    await desiteAPI.resetMaterials(true);
    await desiteAPI.setMaterialToObjects(gOverdueMaterialName, timings.overdueElementList);
    await createAndActivateColorSchema();
    await setCurrentStatus(StatusType.Overdue);
}

async function applyStatusNone()
{
    await desiteAPI.resetMaterials(true);
    await setCurrentStatus(StatusType.None);
}

async function onStatusDropdownMenuChanged()
{
    let dropdown = document.getElementById('dropdownStatus');
    let selectedIndex = dropdown.selectedIndex;
    let statusType = dropdown.options[selectedIndex].value;
    if(statusType == StatusType.None)
        applyStatusNone();
    else if(statusType == StatusType.All)
        applyStatusAll();
    else if(statusType == StatusType.Overdue)
        applyStatusOverdue();
    else if(statusType == StatusType.OnTime)
        applyStatusOnTime();
}

async function onShowOnlyStatusClicked()
{
    gShowOnlyStatus = $("#showOnlyStatusCheckbox").is(":checked");
    showElementsBasedOnFilter();
}

async function onShowAll()
{
    document.getElementById('dropdownLevel').selectedIndex = 0;
    gSelectedLevelName = UNDEFINED_OPTION;
    document.getElementById('dropdownCategory').selectedIndex = 0;    
    gSelectedCategoryName = UNDEFINED_OPTION;
    document.getElementById('showOnlyStatusCheckbox').checked = false;
    gShowOnlyStatus = false;
    showElementsBasedOnFilter();
}

async function onDateChanged()
{
    let selectedElementList = await desiteAPI.getSelectedElements();
    if (selectedElementList.length != 1)
        return;
    let firstSelectedElement = selectedElementList[0];
    let linkedActivityList = await desiteAPI.getLinkedObjects(firstSelectedElement ,'activities', 'geometry');
    if (linkedActivityList.length == 0)
        return;
    firstLinkedActivity = linkedActivityList[0];
    if ($('#inputDateStart').val())
    {
        let startDate = convertDatepickerFormatToUTCDate($('#inputDateStart').val(), 'T08:00:00');
        await desiteAPI.setPropertyValue(firstLinkedActivity, "Start", "xs:dateTime", startDate );
    }
    if ($('#inputDateFinish').val())
    {
        let finishDate = convertDatepickerFormatToUTCDate($('#inputDateFinish').val(), 'T17:00:00');
        await desiteAPI.setPropertyValue(firstLinkedActivity, "Finish", "xs:dateTime", finishDate);
    }
    if ($('#inputDateActualFinish').val())
    {
        let actualFinishDate = convertDatepickerFormatToUTCDate($('#inputDateActualFinish').val(), 'T17:00:00');
        await desiteAPI.setPropertyValue(firstLinkedActivity, "ActualFinish", "xs:dateTime", actualFinishDate);
    }
    if (gCurrentStatus == StatusType.All)
        applyStatusAll()
    else if (gCurrentStatus == StatusType.OnTime)
        applyStatusOnTime();
    else if (gCurrentStatus == StatusType.Overdue)
        applyStatusOverdue();
}

async function createActivityHierarchyNameList(activityElement)
{
    let hierarchyNameList = [];
    let activityName = await desiteAPI.getPropertyValue(activityElement, "cpName", "xs:string");
    if (activityName != undefined)
    {
        let parentActivityElement = await desiteAPI.getParent(activityElement);
        if (parentActivityElement)
            hierarchyNameList = hierarchyNameList.concat(await createActivityHierarchyNameList(parentActivityElement));
        hierarchyNameList.push(activityName);
    }
    return hierarchyNameList;
}

async function createExcelTable(activityElementList)
{
    let currentDate = Date();
    let currentDateInDatepickerFormat = convertUTCDateToDatepickerFormat(currentDate);
    let userName = await desiteAPI.getUserName();
    let companyName = "Your Company";
    let projectNumber = "A001";
    let createdBy = "Implemented by DESITE BIM Team &copy";
 
    let excelTable =
    `<table style="font-size: 24; font-family: Arial, Helvetica, sans-serif; table-layout: fixed; margin: 4px" cellpadding="4" cellspacing="2" width="100%">
    <tr><td style="background:#ACC22B" colspan="15">
    <tr><td colspan="2">Company: ${companyName}</td>
    <td>User name: ${userName}</td>
    <td>Date: ${currentDateInDatepickerFormat}</td></tr>
    <tr><td colspan="2">Project number: ${projectNumber}</td></tr>
    <tr><td colspan="15"><hr></td></tr>
    <tr style="font-weight: bold;">
    <td style="background: #f0eeee; text-align: center; border-bottom: 5px;" colspan="1"></td>
    <td style="background: #f0eeee; text-align: center; border-bottom: 5px;" colspan="6">Activity name</td>
    <td style="background: #f0eeee; text-align: center; border-bottom: 5px;">Finish date</td>
    <td style="background: #f0eeee; text-align: center; border-bottom: 5px;">Actual Finish date</td>
    <td style="background: #f0eeee; text-align: center; border-bottom: 5px; width:auto;">Time difference in days</td>
    <td style="background: #f0eeee; text-align: center; border-bottom: 5px;">Status</td>
    </tr>`;

    for (i in activityElementList)
    {
        let activityElement = activityElementList[i];
        let hasNoLinkedObjects = (await desiteAPI.getPropertyValue(activityElement, "cpCountLinkedObjects", "xs:int") == undefined);
        if (hasNoLinkedObjects)
            continue;        
        let finishDate = await desiteAPI.getPropertyValue(activityElement, "Finish", "xs:dateTime");
        let actualFinishDate = await desiteAPI.getPropertyValue(activityElement, "ActualFinish", "xs:dateTime");        
        let finishDateInDatepickerFormat = finishDate ? convertUTCDateToDatepickerFormat(finishDate) : '';     
        let actualFinishDateInDatepickerFormat = actualFinishDate ? convertUTCDateToDatepickerFormat(actualFinishDate) : '';        
        let differenceFinishAndActualFinishInDays = 0;
        if (finishDate && actualFinishDate)
        {
            let convertMillisecondsToDays = 1/(1000 * 60 * 60 * 24);        
            differenceFinishAndActualFinishInDays = convertMillisecondsToDays*(new Date(actualFinishDate) - new Date(finishDate));                    
        }                
        let activityHierarchy = '';
        let hierarchyNameList = await createActivityHierarchyNameList(activityElement);
        for (j in hierarchyNameList)
            activityHierarchy += (j == (hierarchyNameList.length - 1) ? hierarchyNameList[j] : hierarchyNameList[j] + ', ');        
        let isOnTime = (differenceFinishAndActualFinishInDays <= 0);
        let status = (isOnTime ? 'On time' : 'Overdue');
        let statusColor = (isOnTime ? '#008200' : '#c80000');

        excelTable +=
        `<tr>
        <td width='200'/>
        <td colspan="6">${activityHierarchy}</td>
        <td style='text-align:center'>${finishDateInDatepickerFormat}</td>
        <td style='text-align:center'>${actualFinishDateInDatepickerFormat}</td>
        <td style='text-align:center'>${differenceFinishAndActualFinishInDays.toFixed(2)}</td>
        <td style='color:${statusColor}; text-align:center'>${status}</td>
        </tr>`;
    }

    excelTable += 
    `<tr/>
    <tr/>
    <tr><td style="background:#ddd" colspan="15"><i>${createdBy}</i>
    <tr><td style="background:#ACC22B" colspan="15">
    </table>`;
                                                                                                                    
    return excelTable;
}

async function onTableExportClicked()
{    
    let activityElementList = await desiteAPI.getAllElements("activities"); 
    let excelTable = await createExcelTable(activityElementList);
    let filePath = await desiteAPI.getProjectDirectory() + "/StatusReport.xls";
    await desiteAPI.writeFile(filePath, excelTable);
    await desiteAPI.openLink('file:///' + filePath, true);
}